################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from py_fast_trie.x_fast import TrieNode


def test_init():
    left = TrieNode("left", True)
    right = TrieNode("right", True)
    par = TrieNode("parent", False, left, right)

    assert left.value == "left"
    assert left.leaf is True
    assert left.left is None
    assert left.right is None
    assert left.parent is None

    assert right.value == "right"
    assert right.leaf is True
    assert right.left is None
    assert right.right is None
    assert right.parent is None

    assert par.value == "parent"
    assert par.leaf is False
    assert par.left is left
    assert par.right is right
    assert par.parent is None



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################


from functools import partial
from sys import maxsize

from hypothesis import settings
from hypothesis.strategies import binary, integers, lists, none, one_of

from py_fast_trie import XFastTrie


max_trie_entry_size = (
    maxsize.bit_length() + 1 if settings._current_profile == "ci" else 24
)
max_trie_value = (2**max_trie_entry_size) - 1

to_int = partial(XFastTrie._to_int, length=max_trie_entry_size)

invalid_binary_entry = binary(min_size=(max_trie_entry_size // 8 + 1))
invalid_int_entry = one_of(
    integers(max_value=-1), integers(min_value=(max_trie_value + 2))
)
invalid_trie_entry = one_of(invalid_binary_entry, invalid_int_entry, none())

valid_binary_entry = binary(min_size=1, max_size=(max_trie_entry_size // 8))
valid_int_entry = integers(min_value=0, max_value=max_trie_value)
valid_int_entries = lists(
    valid_int_entry, min_size=1, max_size=max_trie_value, unique=True
)
valid_trie_entry = one_of(valid_binary_entry, valid_int_entry)
valid_trie_entries = lists(
    valid_trie_entry, min_size=1, max_size=max_trie_value, unique_by=to_int
)



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from __future__ import annotations

from collections.abc import Iterable
from sys import maxsize
from typing import Optional, cast

from py_hopscotch_dict import HopscotchDict
from sortedcontainers import SortedList

from py_fast_trie import XFastTrie
from py_fast_trie.x_fast import TrieNode


class YFastTrie:
    @staticmethod
    def _calculate_representative(value: int, max_length: int) -> int:
        """
        Calculate the smallest value that would exist in _partitions
        the given value could belong to

        :param val: The value to calculate the representative for
        :param max_length: The bit length of the largest possible representative

        :return: The closest possible representative to the given number
        """
        result = min(
            max_length * (value // max_length) + (-1 % max_length), 2**max_length - 1
        )
        return cast(int, result)

    @staticmethod
    def _merge_subtrees(
        left_tree: SortedList, right_tree: SortedList, max_size: int
    ) -> tuple[SortedList, Optional[SortedList]]:
        """
        Combine the elements of two trees into one larger tree,
        splitting them again if the larger tree exceeds a given size

        :param left_tree: Tree containing smaller elements
        :param right_tree: Tree containing larger elements
        :param max_size: Maximum size the combined tree can be before splitting
        :return: The combined tree and None if both trees' elements are
                         less than max_size, the tree with the smaller elements
                         and the tree with the larger elements otherwise
        """
        if len(left_tree) + len(right_tree) <= max_size:
            left_tree.update(right_tree)
            result = (left_tree, None)
        else:
            total_median = (len(left_tree) + len(right_tree)) // 2 - 1
            if total_median < len(left_tree):
                total_median = left_tree[total_median]
            else:
                total_median = right_tree[total_median - len(left_tree)]

            # Taking advantage of the fact the number of elements in a subtree can only
            # be twice the bit length of the maximum element before splitting
            median_rep = YFastTrie._calculate_representative(
                total_median, max_size // 2
            )

            if median_rep <= max(left_tree):
                from_tree = left_tree
                to_tree = right_tree
                side = -1
            else:
                from_tree = right_tree
                to_tree = left_tree
                side = 0

            while max(left_tree) > median_rep or min(right_tree) <= median_rep:
                to_tree.add(from_tree.pop(side))

            result = (left_tree, right_tree)

        return result

    @staticmethod
    def _split_subtree(
        tree: SortedList, max_length: int
    ) -> tuple[SortedList, SortedList]:
        """
        Split a tree by its median element into two smaller trees

        :param tree: The tree to split
        :param max_length: The size of the largest possible element in the trie in bits
        :return: The tree with the smaller elements,
                         and the tree with the larger elements
        """
        median = tree.bisect_right(
            YFastTrie._calculate_representative(tree[len(tree) // 2], max_length)
        )
        return SortedList(tree.islice(stop=median)), SortedList(
            tree.islice(start=median)
        )

    def clear(self) -> None:
        """
        Remove all values from the trie and return it to its starting state
        """
        self._count = 0
        self._max: Optional[int] = None
        self._min: Optional[int] = None
        self._partitions = XFastTrie(self._maxlen)
        self._subtrees = HopscotchDict()

    def _get_value_subtree(
        self, value: int, create_subtree: bool = False
    ) -> tuple[Optional[SortedList], Optional[TrieNode]]:
        """
        Find the subtree that would hold the given value

        :param value: The value to find
        :param create_subtree: If there is no subtree that would hold the given value,
                               create one
        :return: The subtree that potentially holds the given value,
                         and its corresponding representative
        """
        result = None

        if self._count == 0:
            rep_node = None
        elif value <= cast(int, self._min) or self._min is None:
            rep_node = self._partitions._min
        else:
            # As the X-fast trie looks for strict successors,
            # if the value being searched for is a representative,
            # the wrong representative will be returned if the one
            # being searched for is not the largest,
            # and no representative will be returned at all if it is;
            # so subtract one before searching for the successor
            rep_node = self._partitions.successor(value - 1)

        if rep_node is None:
            if create_subtree:
                rep = self._calculate_representative(value, self._maxlen)
                self._partitions += rep
                rep_node = self._partitions.successor(rep - 1)
                self._subtrees[rep] = result = SortedList()
        else:
            # Every representative in the X-fast trie
            # should have a corresponding SortedList;
            # the code should blow up if it doesn't
            result = self._subtrees[rep_node.value]

        return (result, rep_node)

    def insert(self, value: int | bytes) -> None:
        """
        Insert a value into the trie

        :param value: The value to insert into the trie
        """
        value = XFastTrie._to_int(value, self._maxlen)
        subtree: SortedList
        subtree, rep_node = self._get_value_subtree(value, True)
        rep_node = cast(TrieNode, rep_node)
        # Do nothing if the value is already in the trie
        if value in subtree:
            return

        if self._max is None or value > self._max:
            self._max = value

        if self._min is None or value < self._min:
            self._min = value

        subtree.add(value)

        if len(subtree) > self._max_subtree_size:
            # Out with the old
            del self._subtrees[rep_node.value]
            self._partitions -= cast(int, rep_node.value)

            # In with the new
            for tree in self._split_subtree(subtree, self._maxlen):
                rep = self._calculate_representative(max(tree), self._maxlen)
                self._partitions += rep
                self._subtrees[rep] = tree

        self._count += 1

    def predecessor(self, value: int | bytes) -> Optional[int]:
        """
        Find the largest value in the trie strictly less than the given value,
        if it exists

        :param value: The value to find the predecessor of
        :return: The predecessor of the given value, or None if it doesn't exist
        """
        value = XFastTrie._to_int(value, self._maxlen)
        subtree, rep_node = self._get_value_subtree(value)

        # subtree should be None only if the trie is empty
        if subtree is None and self._count == 0:
            raise ValueError("No values exist in trie")
        elif value <= cast(int, self._min) or self._min is None:
            return None
        elif value > cast(int, self._max):
            return self._max

        subtree = cast(SortedList, subtree)
        rep_node = cast(TrieNode, rep_node)
        if min(subtree) >= value:
            subtree = self._subtrees[cast(TrieNode, rep_node.left).value]

        return cast(int, subtree[subtree.bisect_left(value) - 1])

    def remove(self, value: int | bytes) -> None:
        """
        Remove the given value from the trie

        :param value: The value to remove from the trie
        """
        value = XFastTrie._to_int(value, self._maxlen)
        subtree, rep_node = self._get_value_subtree(value)

        # There should be no subtree only if the given value is not in the trie
        if subtree is None or value not in subtree:
            raise ValueError("Value does not exist in trie")

        subtree = cast(SortedList, subtree)
        rep_node = cast(TrieNode, rep_node)
        if self._min == value:
            if len(subtree) > 1:
                min_succ = subtree[1]
            else:
                min_succ = self.successor(value)
        else:
            min_succ = -1

        if self._max == value:
            if len(subtree) > 1:
                max_pred = subtree[-2]
            else:
                max_pred = self.predecessor(value)
        else:
            max_pred = -1

        if min_succ != -1:
            self._min = min_succ

        if max_pred != -1:
            self._max = max_pred

        subtree.remove(value)

        if len(subtree) == 0:
            del self._subtrees[rep_node.value]
            self._partitions -= cast(int, rep_node.value)

        elif len(subtree) < self._min_subtree_size and len(self._partitions) > 1:
            if rep_node.left is not None:
                left_rep = rep_node.left
                right_rep = rep_node
            else:
                left_rep = rep_node
                right_rep = cast(TrieNode, rep_node.right)

            left_tree = self._subtrees[left_rep.value]
            right_tree = self._subtrees[right_rep.value]

            # Out with the old
            del self._subtrees[left_rep.value]
            del self._subtrees[right_rep.value]
            self._partitions -= cast(int, left_rep.value)
            self._partitions -= cast(int, right_rep.value)

            # In with the new
            tree: SortedList
            for tree in filter(
                None, self._merge_subtrees(left_tree, right_tree, 2 * self._maxlen)
            ):
                rep = self._calculate_representative(max(tree), self._maxlen)
                self._partitions += rep
                self._subtrees[rep] = tree

        self._count -= 1

    def successor(self, value: int | bytes) -> Optional[int]:
        """
        Find the smallest value in the trie strictly greater than the given value,
        if it exists

        :param value: The value to find the successor of
        :return: The successor of the given value, or None if it doesn't exist
        """
        value = XFastTrie._to_int(value, self._maxlen)
        subtree, rep_node = self._get_value_subtree(value)

        # subtree should be None only if the trie is empty
        if subtree is None and self._count == 0:
            raise ValueError("No values exist in trie")
        elif value >= cast(int, self._max) or self._max is None:
            return None
        elif value < cast(int, self._min):
            return self._min

        subtree = cast(SortedList, subtree)
        rep_node = cast(TrieNode, rep_node)
        if max(subtree) <= value:
            subtree = self._subtrees[cast(TrieNode, rep_node.right).value]

        return cast(int, subtree[subtree.bisect_right(value)])

    @property
    def max(self) -> Optional[int]:
        """
        The maximum value in the trie

        :return: The maximum value in the trie,
                         or None if the trie is empty
        """
        return self._max

    @property
    def min(self) -> Optional[int]:
        """
        The minimum value in the trie

        :return: The minimum value in the trie,
                         or None if the trie is empty
        """
        return self._min

    def __init__(self, max_length: int = (maxsize.bit_length() + 1)):
        self._maxlen = max_length
        self._min_subtree_size = max_length // 2
        self._max_subtree_size = max_length * 2
        self.clear()

    def __contains__(self, value: int | bytes) -> bool:
        value = XFastTrie._to_int(value, self._maxlen)
        subtree, _ = self._get_value_subtree(value)
        return subtree is not None and value in subtree

    def __gt__(self, value: int | bytes) -> Optional[int]:
        return self.successor(value)

    def __iadd__(self, value: int | bytes) -> YFastTrie:
        self.insert(value)
        return self

    def __isub__(self, value: int | bytes) -> YFastTrie:
        self.remove(value)
        return self

    def __iter__(self) -> Iterable[int]:
        for rep in cast(Iterable[int], sorted(self._subtrees)):
            yield from self._subtrees[rep]

    def __len__(self) -> int:
        return self._count

    def __lt__(self, value: int | bytes) -> Optional[int]:
        return self.predecessor(value)



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from hypothesis import HealthCheck, settings
from hypothesis.database import ExampleDatabase


settings.register_profile(
    "ci",
    database=ExampleDatabase(":memory:"),
    deadline=None,
    max_examples=500,
    stateful_step_count=300,
    suppress_health_check=[HealthCheck.too_slow],
)



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from __future__ import annotations

from collections.abc import Iterable
from sys import maxsize
from typing import Optional, cast

from py_hopscotch_dict import HopscotchDict


class TrieNode:
    def __init__(
        self,
        value: Optional[int],
        leaf: bool,
        left: Optional[TrieNode] = None,
        right: Optional[TrieNode] = None,
    ) -> None:
        self.leaf = leaf
        self.value = value
        self.left = left
        self.right = right
        self.parent: Optional[TrieNode] = None


class XFastTrie:
    @staticmethod
    def _make_level_tables(levels: int) -> list[HopscotchDict]:
        """
        Creates the dicts used when searching for a value in the trie

        :param levels: The number of levels in the trie
        :return: Search structures for each level of the trie
        """
        return [HopscotchDict() for _ in range(levels)]

    @staticmethod
    def _to_int(value: int | bytes, length: int) -> int:
        """
        Confirm the desired value could be contained in the table,
        then perform any necessary conversions to the canonical value format

        :param value: The value to be converted
        :param length: The maximum bit length of a value in the trie
        :return: The value converted to an int
        """
        if isinstance(value, int):
            if value.bit_length() > length:
                raise ValueError("Value is too big to be stored in trie")
            elif value < 0:
                raise ValueError("Negative values cannot be stored in trie")
            else:
                return value

        elif isinstance(value, bytes):
            if len(value) * 8 > length:
                raise ValueError("Value is too big to be stored in trie")

            else:
                return sum(map(lambda t: t[1] << 8 * t[0], enumerate(reversed(value))))

        else:
            raise TypeError("Only integers and byte sequences can be stored in trie")

    def clear(self) -> None:
        """
        Empty the trie of all values
        """
        self._count = 0
        self._level_tables = self._make_level_tables(self._maxlen)
        self._min: Optional[TrieNode] = None
        self._max: Optional[TrieNode] = None
        self._root = TrieNode(None, False)

    def _get_closest_ancestor(self, value: int) -> tuple[TrieNode, int]:
        """
        Find the node in the trie with the longest prefix that matches the given value

        :param value: The value to search for
        :return: The node with the longest prefix matching the given value,
                 and its depth in the trie
        """
        result = self._root
        # Moving the root node out of the level tables allows for storing n-bit integers
        result_level = -1

        low_side = 0
        high_side = self._maxlen - 1

        while low_side <= high_side:
            level = (low_side + high_side) // 2
            prefix = value >> (self._maxlen - level - 1)

            if prefix not in self._level_tables[level]:
                high_side = level - 1
            else:
                result = self._level_tables[level][prefix]
                result_level = level
                low_side = level + 1

        return (result, result_level)

    def _get_closest_leaf(self, value: int) -> Optional[TrieNode]:
        """
        Find the leaf in the trie with the value closest to the given value

        :param value: The value to search for
        :return: The leaf with the closest value to the given value
        """
        result = None
        ancestor, level = self._get_closest_ancestor(value)

        # The value is stored in the trie and therefore is the closest leaf to itself
        if ancestor.leaf:
            return ancestor

        # The pointer to the next level down is a descendant pointer
        # and may be larger or smaller than the given value, depending on the leg
        else:
            leg = bin(value)[2:].zfill(self._maxlen)[level + 1]
            descendant = ancestor.left if leg == "0" else ancestor.right

            if descendant is not None:
                # The descendant pointer only points to values
                # on the other leg of the ancestor;
                # make sure there is no leaf not a child of the ancestor which is closer
                candidate = descendant.left if leg == "0" else descendant.right

                if candidate is None or abs(cast(int, descendant.value) - value) < abs(
                    cast(int, candidate.value) - value
                ):
                    result = descendant
                else:
                    result = candidate
        return result

    def insert(self, value: int | bytes) -> None:
        """
        Add the given value to the trie

        :param value: The value to add to the trie
        """
        value = self._to_int(value, self._maxlen)

        # Do nothing if the value is already in the trie
        if value in self._level_tables[-1]:
            return

        leaf_pred = leaf_succ = None

        if self._count > 0:
            neighbor = cast(TrieNode, self._get_closest_leaf(value))

            # The closest leaf is the predecessor of the value to be inserted
            if cast(int, neighbor.value) < value:
                leaf_pred = neighbor
                leaf_succ = neighbor.right

            # The closest leaf is the successor of the value to be inserted
            elif cast(int, neighbor.value) > value:
                leaf_succ = neighbor
                leaf_pred = neighbor.left

        leaf_node = TrieNode(value, True, leaf_pred, leaf_succ)
        value_bits = bin(value)[2:].zfill(self._maxlen)

        # Get ancestor node for new intermediate nodes
        curr_node: Optional[TrieNode]
        depth: int
        curr_node, depth = self._get_closest_ancestor(value)

        for level in range(depth, self._maxlen - 1):
            desc_value = int(value_bits[: level + 2], 2)
            desc_leg = value_bits[level + 1]

            # Inserting leaf node
            if level == self._maxlen - 2:
                descendant = leaf_node

                # Update global min/max pointers as necessary
                if self._min is None or value < cast(int, self._min.value):
                    self._min = descendant

                if self._max is None or value > cast(int, self._max.value):
                    self._max = descendant

                # Wire the new leaf into the linked list at the bottom
                if leaf_pred is not None:
                    leaf_pred.right = descendant

                if leaf_succ is not None:
                    leaf_succ.left = descendant

            # Inserting intermediate node
            else:
                descendant = TrieNode(desc_value, False, None, None)

            # Point the descendant node back to its parent
            descendant.parent = curr_node

            # Point the parent at its new descendant node,
            # making descendant pointers to the new leaf as necessary
            if desc_leg == "0":
                curr_node.left = descendant

                if curr_node.right is None:
                    curr_node.right = leaf_node
            else:
                curr_node.right = descendant

                if curr_node.left is None:
                    curr_node.left = leaf_node

            self._level_tables[level + 1][desc_value] = descendant
            curr_node = descendant

        # Walk up the trie from the leaf node,
        # updating descendant pointers as necessary
        while curr_node is not None:
            if not curr_node.leaf:
                if curr_node.left.leaf and curr_node.left.value > value:  # type: ignore
                    curr_node.left = leaf_node
                elif curr_node.right.leaf and curr_node.right.value < value:  # type: ignore # noqa
                    curr_node.right = leaf_node
            curr_node = curr_node.parent

        self._count += 1

    def predecessor(self, value: int | bytes) -> Optional[TrieNode]:
        """
        Find the largest value in the trie strictly less than the given value

        :param value: The value to find the predecessor for
        :return: The leaf with the largest value strictly less than the given value,
                 or None if the value is at most the value of the smallest leaf
        """
        value = self._to_int(value, self._maxlen)
        node = self._get_closest_leaf(value)

        # This should only happen if there are no values in the trie,
        # But if it could also happen because of some unconsidered edge case,
        # make some noise so the edge case can be fixed
        if node is None:
            raise ValueError("No values exist in trie")
        else:
            return node.left if cast(int, node.value) >= value else node

    def remove(self, value: int | bytes) -> None:
        """
        Remove the given value from the trie

        :param value: The value to remove from the trie
        """
        value = self._to_int(value, self._maxlen)
        curr_node = self._level_tables[-1].get(value)

        # Error when trying to remove a value that hasn't been added
        if curr_node is None:
            raise ValueError("Value does not exist in trie")
        else:
            leaf_pred = curr_node.left
            leaf_succ = curr_node.right
            curr_level = self._maxlen - 1

            # Walk up the trie from the leaf node,
            # modifying/removing internal nodes as necessary
            while curr_node is not None:
                parent = curr_node.parent
                should_delete = False

                # Removing leaf node
                if curr_node.leaf:
                    should_delete = True

                    # Remove the node from the linked list
                    if leaf_pred is not None:
                        leaf_pred.right = leaf_succ

                    if leaf_succ is not None:
                        leaf_succ.left = leaf_pred

                # Root updates
                elif parent is None:
                    if cast(TrieNode, self._min).value == value:
                        self._min = leaf_succ

                    if cast(TrieNode, self._max).value == value:
                        self._max = leaf_pred

                # Removing intermediate nodes
                else:
                    valid_left = valid_right = False

                    if curr_node.left is not None:
                        ptr_level = -1 if curr_node.left.leaf else curr_level + 1
                        if curr_node.left.value in self._level_tables[ptr_level]:
                            valid_left = True

                    if curr_node.right is not None:
                        ptr_level = -1 if curr_node.right.leaf else curr_level + 1
                        if curr_node.right.value in self._level_tables[ptr_level]:
                            valid_right = True

                    should_delete = not (valid_left or valid_right)

                # Trim pointers between the target node at the start of this loop,
                # its parent and its descendants
                if should_delete:
                    node_bits = bin(curr_node.value)[2:].zfill(curr_level + 1)
                    if node_bits[-1] == "0":
                        parent.left = None

                        if parent.right == curr_node:
                            parent.right = None
                    else:
                        parent.right = None

                        if parent.left == curr_node:
                            parent.left = None

                    del self._level_tables[curr_level][curr_node.value]
                    curr_node.left = curr_node.right = curr_node.parent = None
                    del curr_node

                # Update dangling descendant pointers for remaining intermediate nodes
                else:
                    if curr_node.left is None or (
                        curr_node.left.leaf and curr_node.left.value == value
                    ):
                        curr_node.left = leaf_succ

                    if curr_node.right is None or (
                        curr_node.right.leaf and curr_node.right.value == value
                    ):
                        curr_node.right = leaf_pred

                curr_node = parent
                curr_level -= 1

        self._count -= 1

    def successor(self, value: int | bytes) -> Optional[TrieNode]:
        """
        Find the smallest value in the trie strictly greater than the given value

        :param value: The value to find the successor for
        :return: The leaf with the smallest value strictly greater than the given value,
                 or None if the value is at least the value of the largest leaf
        """
        value = self._to_int(value, self._maxlen)
        node = self._get_closest_leaf(value)

        # This should only happen if there are no values in the trie,
        # But if it could also happen because of some unconsidered edge case,
        # make some noise so the edge case can be fixed
        if node is None:
            raise ValueError("No values exist in trie")
        else:
            return node.right if cast(int, node.value) <= value else node

    @property
    def max(self) -> Optional[int]:
        """
        The maximum value in the trie

        :return: The maximum value in the trie,
                 or None if the trie is empty
        """
        return self._max.value if self._max is not None else self._max

    @property
    def min(self) -> Optional[int]:
        """
        The minimum value in the trie

        :return: The minimum value in the trie,
                 or None if the trie is empty
        """
        return self._min.value if self._min is not None else self._min

    def __init__(self, max_length: int = (maxsize.bit_length() + 1)) -> None:
        self._maxlen = max_length
        self.clear()

    def __contains__(self, value: int | bytes) -> bool:
        value = self._to_int(value, self._maxlen)
        return value in self._level_tables[-1]

    def __gt__(self, value: int | bytes) -> Optional[int]:
        result = self.successor(value)
        return result.value if result is not None else result

    def __iadd__(self, value: int | bytes) -> XFastTrie:
        self.insert(value)
        return self

    def __isub__(self, value: int | bytes) -> XFastTrie:
        self.remove(value)
        return self

    def __iter__(self) -> Iterable[int]:
        node = self._min
        while node is not None:
            yield cast(int, node.value)
            node = node.right

    def __len__(self) -> int:
        return self._count

    def __lt__(self, value: int | bytes) -> Optional[int]:
        result = self.predecessor(value)
        return result.value if result is not None else result



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################


from numbers import Integral
from struct import pack, unpack
from sys import maxsize
from test import (
    invalid_trie_entry,
    max_trie_entry_size,
    max_trie_value,
    valid_int_entries,
    valid_int_entry,
    valid_trie_entries,
    valid_trie_entry,
)

import pytest

from hypothesis import given, settings
from hypothesis.stateful import RuleBasedStateMachine, invariant, rule
from hypothesis.strategies import integers, lists

from py_fast_trie import XFastTrie


def to_bytes(val):
    if val.bit_length() < 9:
        fmt = "B"
    elif val.bit_length() < 17:
        fmt = ">H"
    elif val.bit_length() < 33:
        fmt = ">L"
    else:
        fmt = ">Q"

    return pack(fmt, val)


@given(integers(min_value=0, max_value=max_trie_entry_size))
def test_make_level_tables(depth):
    assert len(XFastTrie._make_level_tables(depth)) == depth


@given(valid_trie_entry)
def test_to_int(value):
    if isinstance(value, int):
        assert XFastTrie._to_int(value, max_trie_entry_size) == value

    elif isinstance(value, bytes):
        value_int = unpack(">Q", value.rjust((maxsize.bit_length() + 1) // 8, b"\x00"))[
            0
        ]
        assert XFastTrie._to_int(value, max_trie_entry_size) == value_int


@given(invalid_trie_entry)
def test_to_int_exceptions(value):
    if isinstance(value, Integral):
        with pytest.raises(ValueError):
            XFastTrie._to_int(value, max_trie_entry_size)

    elif isinstance(value, bytes):
        with pytest.raises(ValueError):
            XFastTrie._to_int(value, max_trie_entry_size)

    else:
        with pytest.raises(TypeError):
            XFastTrie._to_int(value, max_trie_entry_size)


@given(valid_trie_entries, valid_int_entries)
def test_get_closest_ancestor(entries, test_values):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    entries = [t._to_int(e, t._maxlen) for e in entries]

    for val in test_values:
        ancestor, level = t._get_closest_ancestor(val)

        if val in entries:
            assert ancestor.leaf
            assert ancestor.value == val

        else:
            assert not ancestor.leaf
            test_bits = bin(val)[2:].zfill(t._maxlen)[: level + 2]

            left_desc_lvl = t._maxlen if ancestor.left.leaf else level + 2
            assert (
                not bin(ancestor.left.value)[2:]
                .zfill(left_desc_lvl)
                .startswith(test_bits)
            )

            right_desc_lvl = t._maxlen if ancestor.right.leaf else level + 2
            assert (
                not bin(ancestor.right.value)[2:]
                .zfill(right_desc_lvl)
                .startswith(test_bits)
            )


@given(valid_trie_entries, valid_int_entries)
def test_get_closest_leaf(entries, test_values):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    entries = [t._to_int(e, t._maxlen) for e in entries]

    for val in test_values:
        neighbor = t._get_closest_leaf(val)
        assert neighbor.leaf

        if val in entries:
            assert neighbor.value == val

        else:
            if neighbor.left is not None:
                assert abs(neighbor.value - val) <= abs(neighbor.left.value - val)

            if neighbor.right is not None:
                assert abs(neighbor.value - val) <= abs(neighbor.right.value - val)


@given(valid_trie_entries, valid_int_entries)
def test_predecessor(entries, test_values):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    for val in test_values:
        pred = t < val

        if pred is not None:
            assert pred < val
            pred = t.predecessor(val)

            if pred.right is not None:
                assert pred.right.value >= val


@given(valid_trie_entries, valid_int_entries)
def test_successor(entries, test_values):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    for val in test_values:
        succ = t > val

        if succ is not None:
            assert succ > val
            succ = t.successor(val)

            if succ.left is not None:
                assert succ.left.value <= val


def test_successor_predecessor_empty_trie():
    t = XFastTrie(max_trie_entry_size)

    with pytest.raises(ValueError):
        t.successor(0)

    with pytest.raises(ValueError):
        t.predecessor(0)


@given(valid_trie_entries)
def test_clear(entries):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    assert len(t) > 0
    assert t._min is not None
    assert t._max is not None

    t.clear()

    for d in t._level_tables:
        assert len(d) == 0

    assert len(t) == 0
    assert t._min is None
    assert t._max is None


@given(lists(valid_int_entry, min_size=0, max_size=max_trie_value, unique=True))
def test_iter(entries):
    t = XFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    entries = sorted(entries)

    for entry in t:
        assert entry == entries.pop(0)

    assert len(entries) == 0


class XFastStateMachine(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.t = XFastTrie(max_trie_entry_size)

    def teardown(self):
        values = list(self.t._level_tables[-1])

        for val in values:
            self.t -= val

    @invariant()
    def valid_count(self):
        assert len(self.t) == len(self.t._level_tables[-1])

    @invariant()
    def valid_min(self):
        if len(self.t) > 0:
            assert self.t._min.left is None

            for leaf in self.t._level_tables[-1].values():
                assert leaf is self.t._min or self.t.min < leaf.value

    @invariant()
    def valid_max(self):
        if len(self.t) > 0:
            assert self.t._max.right is None

            for leaf in self.t._level_tables[-1].values():
                assert leaf is self.t._max or self.t.max > leaf.value

    @invariant()
    def valid_pointers(self):
        for level, table in enumerate(self.t._level_tables):
            for node in table.values():
                if not node.leaf:
                    left_child_value = node.value << 1 & -2
                    right_child_value = node.value << 1 | 1
                    left_child = self.t._level_tables[level + 1].get(left_child_value)
                    right_child = self.t._level_tables[level + 1].get(right_child_value)

                    # While we're in here, make sure the node should be in the trie
                    assert left_child is not None or right_child is not None

                    if left_child is not None:
                        assert node.left is left_child
                        assert left_child.parent is node
                    else:
                        desc = node.right

                        while not desc.leaf:
                            desc = desc.left

                        assert node.left is desc

                    if right_child is not None:
                        assert node.right is right_child
                        assert right_child.parent is node
                    else:
                        desc = node.left

                        while not desc.leaf:
                            desc = desc.right

                        assert node.right is desc
                else:
                    if node.left is not None:
                        assert node.left.value < node.value

                    if node.right is not None:
                        assert node.right.value > node.value

    @rule(val=valid_trie_entry)
    def insert_value(self, val):
        self.t += val

    @rule(val=valid_trie_entry)
    def remove_value(self, val):
        if val not in self.t:
            try:
                self.t -= val
            except ValueError:
                pass
        else:
            self.t -= val


XFastStateMachine.TestCase.settings = settings(max_examples=50, deadline=None)
test_x_fast_trie = XFastStateMachine.TestCase



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################


from itertools import chain
from random import randint
from test import (
    XFastTrie,
    max_trie_entry_size,
    max_trie_value,
    valid_int_entries,
    valid_int_entry,
    valid_trie_entries,
    valid_trie_entry,
)

import pytest

from hypothesis import given, settings
from hypothesis.stateful import RuleBasedStateMachine, invariant, rule
from hypothesis.strategies import integers, lists
from sortedcontainers import SortedList

from py_fast_trie import YFastTrie


@given(integers(min_value=0, max_value=max_trie_value))
def test_calculate_representative(val):
    rep = YFastTrie._calculate_representative(val, max_trie_entry_size)
    assert rep >= val
    assert (rep == 2**max_trie_entry_size - 1) or (rep + 1) % max_trie_entry_size == 0


@given(
    lists(
        valid_int_entry,
        min_size=2,
        max_size=((4 * max_trie_entry_size) - 1),
        unique=True,
    )
)
def test_merge_subtrees(values):
    values = SortedList(values)
    split = randint(1, len(values) - 1)
    left_tree = SortedList(values.islice(stop=split))
    right_tree = SortedList(values.islice(start=split))
    new_left, new_right = YFastTrie._merge_subtrees(
        left_tree, right_tree, 2 * max_trie_entry_size
    )

    if len(values) <= 2 * max_trie_entry_size:
        assert new_right is None
        assert isinstance(new_left, SortedList)
        assert len(new_left) == len(values)
    else:
        assert isinstance(new_left, SortedList)
        assert isinstance(new_right, SortedList)
        assert len(new_left) + len(new_right) == len(values)
        assert YFastTrie._calculate_representative(
            max(new_left), max_trie_entry_size
        ) < min(new_right)


@given(
    lists(
        valid_int_entry,
        min_size=((2 * max_trie_entry_size) + 1),
        max_size=((4 * max_trie_entry_size) - 1),
        unique=True,
    )
)
def test_merge_large_subtrees(values):
    values = SortedList(values)
    split = len(values) // 2
    left_tree = SortedList(values.islice(stop=split))
    right_tree = SortedList(values.islice(start=split))
    new_left, new_right = YFastTrie._merge_subtrees(
        left_tree, right_tree, 2 * max_trie_entry_size
    )
    assert isinstance(new_left, SortedList)
    assert isinstance(new_right, SortedList)
    assert len(new_left) + len(new_right) == len(values)
    assert YFastTrie._calculate_representative(
        max(new_left), max_trie_entry_size
    ) < min(new_right)

    split += 1
    left_tree = SortedList(values.islice(stop=split))
    right_tree = SortedList(values.islice(start=split))
    new_left, new_right = YFastTrie._merge_subtrees(
        left_tree, right_tree, 2 * max_trie_entry_size
    )
    assert isinstance(new_left, SortedList)
    assert isinstance(new_right, SortedList)
    assert len(new_left) + len(new_right) == len(values)
    assert YFastTrie._calculate_representative(
        max(new_left), max_trie_entry_size
    ) < min(new_right)


@given(
    lists(
        valid_int_entry,
        min_size=((2 * max_trie_entry_size) + 1),
        max_size=((4 * max_trie_entry_size) - 1),
        unique=True,
    )
)
def test_split_subtree(values):
    left_tree, right_tree = YFastTrie._split_subtree(
        SortedList(values), max_trie_entry_size
    )

    assert isinstance(left_tree, SortedList)
    assert isinstance(right_tree, SortedList)
    assert len(left_tree) + len(right_tree) == len(values)
    assert max(left_tree) < min(right_tree)
    assert YFastTrie._calculate_representative(
        max(left_tree), max_trie_entry_size
    ) < min(right_tree)


@given(valid_trie_entries, valid_int_entries)
def test_get_value_subtree(entries, test_values):
    t = YFastTrie(max_trie_entry_size)

    assert t._get_value_subtree(test_values[0]) == (None, None)

    for entry in entries:
        t += entry

    entries = [XFastTrie._to_int(e, t._maxlen) for e in entries]

    for val in test_values:
        tree, rep_node = t._get_value_subtree(val)

        if val > t._partitions.max:
            assert tree is None
            assert rep_node is None

        else:
            assert val <= rep_node.value

            if val in entries:
                assert val in tree
            else:
                assert val not in tree


@given(valid_trie_entries, valid_int_entries)
def test_predecessor(entries, test_values):
    t = YFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    for val in test_values:
        pred = t < val

        if pred is not None:
            assert pred < val

            tree, _ = t._get_value_subtree(val)
            if tree is None or min(tree) >= val:
                tree, _ = t._get_value_subtree(pred)

            assert max(v for v in tree if v < val) == pred

        else:
            assert val <= t.min


@given(valid_trie_entries, valid_int_entries)
def test_successor(entries, test_values):
    t = YFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    for val in test_values:
        succ = t > val

        if succ is not None:
            assert succ > val

            tree, _ = t._get_value_subtree(val)
            if tree is None or max(tree) <= val:
                tree, _ = t._get_value_subtree(succ)

            assert min(v for v in tree if v > val) == succ

        else:
            assert val >= t.max


def test_successor_predecessor_empty_trie():
    t = YFastTrie(max_trie_entry_size)

    with pytest.raises(ValueError):
        t.successor(0)

    with pytest.raises(ValueError):
        t.predecessor(0)


@given(valid_trie_entries)
def test_clear(entries):
    t = YFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    assert len(t) > 0
    assert t.min is not None
    assert t.max is not None

    t.clear()

    assert len(t) == 0
    assert len(t._partitions) == 0
    assert len(t._subtrees) == 0
    assert t.min is None
    assert t.max is None


def test_insert_with_split():
    t = YFastTrie(max_trie_entry_size)

    big_rep = 3 * max_trie_entry_size - 1
    small_rep = 2 * max_trie_entry_size - 1
    for i in range(3 * max_trie_entry_size - 1, max_trie_entry_size - 1, -1):
        t += i

    assert len(t._partitions) == 1
    assert big_rep in t._subtrees
    assert len(t._subtrees[big_rep]) == 2 * max_trie_entry_size

    t += max_trie_entry_size - 1

    assert len(t._partitions) == 2
    assert big_rep in t._subtrees
    assert small_rep in t._subtrees
    assert len(t._subtrees[big_rep]) == max_trie_entry_size
    assert len(t._subtrees[small_rep]) == max_trie_entry_size + 1


@given(lists(valid_int_entry, min_size=0, max_size=max_trie_value, unique=True))
def test_iter(entries):
    t = YFastTrie(max_trie_entry_size)

    for entry in entries:
        t += entry

    entries = sorted(entries)

    for entry in t:
        assert entry == entries.pop(0)

    assert len(entries) == 0


class YFastStateMachine(RuleBasedStateMachine):
    def __init__(self):
        super().__init__()
        self.t = YFastTrie(max_trie_entry_size)

    def teardown(self):
        values = list(chain.from_iterable(self.t._subtrees.values()))

        for val in values:
            self.t -= val

    @invariant()
    def valid_count(self):
        tree_sizes = map(lambda t: len(t), self.t._subtrees.values())
        assert len(self.t) == sum(tree_sizes)

    @invariant()
    def valid_min(self):
        if len(self.t) == 0:
            assert self.t.min is None
        elif len(self.t) == 1:
            assert self.t.min == self.t.max
        else:
            for value in chain.from_iterable(self.t._subtrees.values()):
                assert value is self.t.min or self.t.min < value

    @invariant()
    def valid_max(self):
        if len(self.t) == 0:
            assert self.t.max is None
        elif len(self.t) == 1:
            assert self.t.max == self.t.min
        else:
            for value in chain.from_iterable(self.t._subtrees.values()):
                assert value is self.t.min or self.t.min < value

    @invariant()
    def valid_representatives(self):
        rep = self.t._partitions._min
        while rep is not None:
            pred = rep.left
            succ = rep.right

            max(self.t._subtrees[rep.value]) <= rep.value

            if pred is not None:
                max(self.t._subtrees[pred.value]) < min(self.t._subtrees[rep.value])

            if succ is not None:
                min(self.t._subtrees[succ.value]) > max(self.t._subtrees[rep.value])

            rep = rep.right

    @invariant()
    def valid_subtree_values(self):
        for rep in self.t._subtrees.keys():
            assert rep in self.t._partitions
            assert YFastTrie._calculate_representative(rep, max_trie_entry_size) == rep

    @rule(val=valid_trie_entry)
    def insert_value(self, val):
        self.t += val

    @rule(val=valid_trie_entry)
    def remove_value(self, val):
        if val not in self.t:
            try:
                self.t -= val
            except ValueError:
                pass
        else:
            self.t -= val


YFastStateMachine.TestCase.settings = settings(max_examples=50, deadline=None)
test_y_fast_trie = YFastStateMachine.TestCase



################################################################################
#                                 py-fast-trie                                 #
#          Python library for tries with different grades of fastness          #
#                       (C) 2020, 2022-2023 Jeremy Brown                       #
#                Released under Prosperity Public License 3.0.0                #
################################################################################

from py_fast_trie.x_fast import XFastTrie as XFastTrie  # noqa
from py_fast_trie.y_fast import YFastTrie as YFastTrie  # noqa
